/*
 *         _        ___
 *    /\ / /___    / _ \ /\ /\  _ ___
 *   / // // _ \  / // // // // // _ \
 *  / // // // / / ___// // // // // /
 *  \___//_//_/ /_/   /_/ \___/ \_  /
 *                             \___/
 * 
 *  Compunach UnPlug
 *  Copyright (C) 2009 David Batley <unplug@dbatley.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


/**
 * Things the overlay needs, such as opening the popup window.
 */
UnPlug2Overlay = {
	UnPlug2Overlay : function () {
		this._have_loaded_browser = false;
	},
	
	/**
	 * Run a search (by opening a popup window and giving it the page to work with)
	 */
	run : function () {
		try {
			var data = { tgt_window : this.get_current_window(), gbrowser : getBrowser() };
			window.openDialog("chrome://unplug/content/searchpage.xul", "unplug_window", "chrome,centerscreen", data );
		} catch (e) {
			UnPlug2.log(e);
		}
	},
	
	/**
	 * Function which is called once, when chrome://browser has finished loading.
	 * This removes the items from menus which we don't care about.
	 */
	browser_loaded : function () {
		for each (name in [ "toolsmenu", "contextmenu" ]) {
			try {
				if ( ! UnPlug2.get_pref("overlay." + name, true) ) {
					var el = window.document.getElementById("cn_unplug2_" + name)
					if (!el)
						continue;
					el.parent.removeChild(el);
				}
			} catch(e) {
				UnPlug2.log("Error removing " + name + " because " + e);
			}
		}
	},
	
	/**
	 * Function which is called each time a page is loaded
	 */
	page_loaded : function () {
		if ( !this._have_loaded_browser ) {
			this.browser_loaded();
			this._have_loaded_browser = true;
		}
	},
	
	/**
	 * Returns the window in the curently open tab
	 */
	get_current_window : function () {
		var br = getBrowser();
		var br_tab = br.getBrowserAtIndex(br.mTabContainer.selectedIndex);
		return br_tab.contentWindow;
	},
	
	toString : function () {
		return "<js:UnPlug2Overlay>";
	},
	
	version : 2.0 };

// init
UnPlug2Overlay.UnPlug2Overlay();

window.addEventListener("load", function () { UnPlug2Overlay.page_loaded(); }, false);

